;LinE MONitor - record time and duration of power outages
;ATTiny2313 @ 32768 Hz (external clock), Fuses: low=0xE0, high=0x9F
;
;PORTS
;  PA0   = (unused)
;  PA1   = External 32768 Hz Clock input (XTAL2)
;  PB0   = Key "down"/"next"
;  PB1   = Key "up"/"set"
;  PB2   = (unused)
;  PB3   = (unused)
;  PB4-7 = LCD D4-D7
;  PD0   = LCD Power
;  PD1   = DCF77-Receiver Power
;  PD2   = (unused)
;  PD3   = Line Monitor input (high = line voltage present) (INT1)
;  PD4   = LCD RS
;  PD5   = LCD E
;  PD6   = DCF77 Signal input (ICP)

;VERSIONS
;  2008-06-23  Arne Rossius
;  2008-07-06  Arne Rossius
;    * BUGFIX: hours not reset at midnight
;    * BUGFIX: "Rx" displayed together with "DCF", not when reception enabled

.include "../include/tn2313def.inc"

.equ	MESSAGE_TIME = 25 ;message display duration (in ~40 ms, 25 = ~1 s)
.equ	LCD_TIMEOUT = 60 ;LCD Timeout (in seconds)

.equ	KEYPIN = PINB
.equ	Key_Up = 0
.equ	Key_Down = 1

.equ	DCFPORT = PORTD
.equ	DCFPIN = PIND
.equ	DCF_Power = 1
.equ	DCF_Signal = 6
.equ	DCFInvert = 0 ;inverted output (negative pulses)

.equ	LCDData = PORTB
.equ	LCDCtrl = PORTD
.equ	LCD_Power = 0
.equ	LCD_RS = 4
.equ	LCD_E = 5

.equ	LinePIN = PIND
.equ	LineStatus = 3

.equ	INVALID = 0xFF ;invalid value for packed BCD, EEPROM address
.equ	EMPTY = 0xFF ;value in empty EEPROM location

;registers for DCF77 telegram
;         MSB                                       LSB
;       +-----+-----+-----+-----+-----+-----+-----+-----+   LS = leap second
;dcf1 = | m20 | m10 | m8  | m4  | m2  | m1  |  S  | LS  |   S = start bit (1)
;       +-----+-----+-----+-----+-----+-----+-----+-----+   m = minutes
;dcf2 = | h20 | h10 | h8  | h4  | h2  | h1  | MnP | m40 |   MnP = minutes parity
;       +-----+-----+-----+-----+-----+-----+-----+-----+   h = hours
;dcf3 = | WD1 | D20 | D10 | D8  | D4  | D2  | D1  | HrP |   hours parity
;       +-----+-----+-----+-----+-----+-----+-----+-----+   D = day
;dcf4 = | Y1  | M10 | M8  | M4  | M2  | M1  | WD4 | WD2 |   WD = weekday
;       +-----+-----+-----+-----+-----+-----+-----+-----+   M = month
;dcf5 = | DaP | Y80 | Y40 | Y20 | Y10 | Y8  | Y4  | Y2  |   Y = year
;       +-----+-----+-----+-----+-----+-----+-----+-----+   DaP = date parity
.def	dcf1 = R0
.def	dcf2 = R1
.def	dcf3 = R2
.def	dcf4 = R3
.def	dcf5 = R4
;time/date registers
.def	second = R5
.def	minute = R6
.def	hour = R7
.def	day = R8
.def	month = R9
.def	year = R10
.def	wkday = R11 ;(weekday, 1=MO 7=SU)
;User Interface registers
.def	oldkeys = R12
.def	minval = R13
;backup registers for Z pointer
.def	backupL = R14
.def	backupH = R15
;general use registers
.def	temp = R16
.def	temp2 = R17
.def	count = R18
;DCF77-specific registers
.def	dcfbits = R19 ;received bits
.def	dcfflags = R20
	.equ	ErrBitCount = 0 ;last error: wrong number of bits
	.equ	ErrParity = 1 ;last error: parity error
	.equ	ErrBitLength = 2 ;last error: wrong bit length
	.equ	ErrStartBit = 3 ;last error: start bit was not 1
	.equ	TimeOK = 4 ;time is OK
	.equ	StopDCF = 5 ;stop DCF77 reception until a new telegram starts
	.equ	DCFdisabled = 6 ;DCF77 receiver disabled
	.equ	DCFTime = 7 ;Last synchronisation attempt was successful
	.equ	ErrMask = 1<<ErrBitLength | 1<<ErrParity | 1<<ErrBitCount | 1<<ErrStartBit
;Power Outage registers
.def	pwrsec = R21
.def	pwrmin = R22
.def	pwrhrL = R23
.def	pwrhrH = R24
;User Interface registers
.def	lcdtime = R25
.def	dispaddr = R26
.def	bothkeys = R27
.def	maxval = R28
.def	keys = R29
	.equ	kUp = Key_Up
	.equ	kDown = Key_Down
	.equ	kDownPress = 2
	.equ	kUpPress = 3
	.equ	kBothPress = 4
	.equ	Keypresses = 1<<kDownPress | 1<<kUpPress | 1<<kBothPress

;R31:R30 = Z register for Flash/RAM access
	
;===============================================================================

.eseg
	.dw 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF
	.dw 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF
	.dw 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF
	.dw 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF
	.dw 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF
	.dw 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF
	.dw 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF
	.dw 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF, 0xFFFF

;===============================================================================

.cseg
.org 0x0000
	ldi	temp, LOW(RAMEND)
	rjmp	start
.org INT1addr
	rjmp	extint1
.org ICP1addr
	push	temp
	rjmp	icp1
.org OVF1addr
	;reset timer to 0x0101 (> 2s) [ldi and out don't change SREG]
	push	temp
	ldi	temp, 1
	out	TCNT1H, temp
	out	TCNT1L, temp
	pop	temp
	reti
.org PCINTaddr
	push	temp
	rjmp	pcint
.org OC0Aaddr
	rjmp	ovf0
	
;===============================================================================

weekdays:
	;.db "Mo", "Di", "Mi", "Do", "Fr", "Sa", "So" ;German
	.db "Mo", "Tu", "We", "Th", "Fr", "Sa", "Su" ;English

strStart:
	.db 0,1,2,3,"    LeMon   "
	.db 4,5,6,7,"  v20080622 "
	
strDCFInfo:
	.db "00 Bits, last=? "
	.db "Err:- Stop In:? "
	
strEmpty:
	.db "No power outages"
	.db "  recorded yet. "
	
strStop: .db "Stop",0,0
strDCF: .db "DCF",0
strRx: .db "Rx",0,0
strSet: .db "Set",0
strSpace: .db "    ",0,0

;charLemon: ;lemon outline w/ shadow
;	.db 0x00, 0x00, 0x01, 0x02, 0x04, 0x04, 0x08, 0x08
;	.db 0x00, 0x1f, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01
;	.db 0x00, 0x1c, 0x0c, 0x14, 0x0c, 0x14, 0x0c, 0x14
;	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
;	.db 0x08, 0x08, 0x09, 0x0a, 0x0d, 0x0f, 0x00, 0x00
;	.db 0x01, 0x02, 0x1f, 0x0a, 0x15, 0x1e, 0x00, 0x00
;	.db 0x14, 0x08, 0x18, 0x10, 0x00, 0x00, 0x00, 0x00
;	.db 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00
	
charLemon: ;lemon slice
	.db 0x00, 0x00, 0x00, 0x01, 0x01, 0x02, 0x02, 0x02
	.db 0x07, 0x18, 0x03, 0x07, 0x03, 0x19, 0x1c, 0x1e
	.db 0x1c, 0x03, 0x18, 0x1c, 0x18, 0x13, 0x07, 0x0f
	.db 0x00, 0x00, 0x00, 0x10, 0x10, 0x08, 0x08, 0x08
	.db 0x02, 0x02, 0x02, 0x01, 0x01, 0x00, 0x00, 0x00
	.db 0x1e, 0x1c, 0x19, 0x03, 0x07, 0x03, 0x18, 0x07
	.db 0x0f, 0x07, 0x13, 0x18, 0x1c, 0x18, 0x03, 0x1c
	.db 0x08, 0x08, 0x08, 0x10, 0x10, 0x00, 0x00, 0x00

;===============================================================================

extint1:
	;External Interrupt 1: Power Line state changed (power/no power)
	push	temp
	push	temp2
	push	ZL
	in	temp, SREG
	push	temp
	
	sbic	LinePIN, LineStatus
	rjmp	power_on
	
	;power is out
	cpi	pwrsec, INVALID
	brne	extint1_end ;power outage was already noticed
	;reset power outage duration timer
	ldi	pwrsec, 0x00
	ldi	pwrmin, 0x00
	ldi	pwrhrL, 0x00
	ldi	pwrhrH, 0x00
	;store time and date
	rcall	eeprom_findfree
	mov	temp, year
	rcall	eeprom_write
	mov	temp, month
	rcall	eeprom_write
	mov	temp, day
	rcall	eeprom_write
	mov	temp, hour
	rcall	eeprom_write
	mov	temp, minute
	rcall	eeprom_write
	mov	temp, second
	rcall	eeprom_write
	ldi	temp, EMPTY
	rcall	eeprom_write
	rcall	eeprom_write
	rcall	eeprom_write
	rcall	eeprom_write
	rcall	eeprom_write ;write INVALID to first byte of next block
	rjmp	extint1_end
	
power_on:
	;power is present
	cpi	pwrsec, INVALID
	breq	extint1_end ;power was already present before
	;store power outage duration
	rcall	eeprom_findfree
	subi	ZL, 10
	brcc	PC+2
	ldi	ZL, 110 ;wrap around, last block is at address 110~119
	subi	ZL, -6 ;add offset for duration
	mov	temp, pwrhrH
	rcall	eeprom_write
	mov	temp, pwrhrL
	rcall	eeprom_write
	mov	temp, pwrmin
	rcall	eeprom_write
	mov	temp, pwrsec
	rcall	eeprom_write
	ldi	pwrsec, INVALID ;invalidate old information
	
extint1_end:
	pop	temp
	out	SREG, temp
	pop	ZL
	pop	temp2
	pop	temp
	reti

;===============================================================================

pcint:
	;Pin Change Interrupt: key pressed or released
	;push	temp --> moved to beginning of code segment
	push	temp2
	in	temp, SREG
	push	temp
	
	;read keys
	in	temp, KeyPIN
	com	temp
	andi	temp, 1<<Key_Up | 1<<Key_Down
	
	;reset LCD timeout if a key is being pressed
	cpi	keys, 0x00
	breq	PC+2
	ldi	lcdtime, LCD_TIMEOUT
	
	;set/reset 'bothkeys' duration timer
	cpi	temp, 1<<kUp | 1<<kDown
	brne	pcint_notboth
	cpi	bothkeys, 0
	brne	pcint_bothend
	ldi	bothkeys, 1
	rjmp	pcint_bothend
pcint_notboth:
	ldi	bothkeys, 0
pcint_bothend:

	;check for changed keys (pressed or released since last interrupt)
	eor	oldkeys, temp
	tst	oldkeys
	breq	pcint_keyend ;no change
	
	;set keys register:
	;  bit  6   = set to 1 if both keys pressed down after last interrupt
	;  bits 5/4 = set to 1 if key was pressed down after last interrupt
	;  bits 1/0 = set to 1 if key is being pressed
	andi	keys, Keypresses ;only keep keypress information
	or	keys, temp ;add pressed keys
	and	oldkeys, temp ;add newly pressed keys
	sbrc	oldkeys, kUp
	ori	keys, 1<<kUpPress
	sbrc	oldkeys, kDown
	ori	keys, 1<<kDownPress
	tst	oldkeys ;add both keys pressed flag
	breq	pcint_keyend
	cpi	temp, 1<<kUp | 1<<kDown
	brne	pcint_keyend
	ori	keys, 1<<kBothPress
	
pcint_keyend:
	mov	oldkeys, temp
	
	pop	temp
	out	SREG, temp
	pop	temp2
	pop	temp
	reti

;===============================================================================

ovf0:
	;Timer 0 overflow interrupt (1 Hz)
	push	temp
	push	temp2
	in	temp, SREG
	push	temp

;---------------------------------------
; update timeout and duration counters
;---------------------------------------
	
	;LCD timeout
	cpi	lcdtime, 0
	breq	PC+2
	dec	lcdtime
	
	;"both keys" duration counter
	cpi	bothkeys, 0
	breq	PC+2
	inc	bothkeys
	sbrc	bothkeys, 2 ;stop at bothkeys == 3
	dec	bothkeys

;---------------------------------------
; increment time
;---------------------------------------

	;don't increment if time is still unset
	sbrs	dcfflags, TimeOK
	rjmp	inctime_end

	;increment seconds
	mov	temp, second
	rcall	bcdinc
	mov	second, temp
	cpi	temp, 0x60
	brlo	inctime_end
	
	;increment minutes
	clr	second
	mov	temp, minute
	rcall	bcdinc
	mov	minute, temp
	cpi	temp, 0x10 ;disable DCF77 reception if nothing received for 10 minutes
	brne	inctime_nodisable
	sbrs	dcfflags, TimeOK
	rjmp	inctime_nodisable
	andi	dcfflags, ~(1<<DCFTime)
	ori	dcfflags, 1<<DCFdisabled | 1<<StopDCF
	cbi	DCFPORT, DCF_Power
inctime_nodisable:
	cpi	temp, 0x60
	brlo	inctime_end

	;increment hours
	clr	minute
	mov	temp, hour
	rcall	bcdinc
	mov	hour, temp
	cpi	temp, 0x03
	brne	inctime_noenable
	andi	dcfflags, ~(1<<DCFdisabled) ;re-enable DCF77 reception at 03:00
	sbi	DCFPORT, DCF_Power
inctime_noenable:
	cpi	temp, 0x24
	brlo	inctime_end
	
	;increment weekday and days
	clr	hour
	mov	temp, wkday ;increment weekday
	inc	temp
	sbrc	temp, 3 ;temp == 8 ?
	ldi	temp, 1
	mov	wkday, temp
	mov	temp, day ;increment day
	rcall	bcdinc
	mov	day, temp
	rcall	monthlength
	cp	temp, day
	brsh	inctime_end
	
	;increment month
	clr	day
	inc	day
	mov	temp, month
	rcall	bcdinc
	mov	month, temp
	cpi	temp, 0x13
	brlo	inctime_end
	
	;increment year
	mov	month, day ;day is 0x01
	mov	temp, year
	rcall	bcdinc
	mov	year, temp
inctime_end:

;---------------------------------------
; increment power outage duration
;---------------------------------------

	cpi	pwrsec, INVALID
	breq	incpwr_end ;no power outage
	cpi	pwrhrH, 0x10
	brsh	incpwr_end ;overflow
	
	;second
	mov	temp, pwrsec
	rcall	bcdinc
	mov	pwrsec, temp
	cpi	temp, 0x60
	brlo	incpwr_end
	
	;minute
	clr	pwrsec
	mov	temp, pwrmin
	rcall	bcdinc
	mov	pwrmin, temp
	cpi	temp, 0x60
	brlo	incpwr_end
	
	;hour
	clr	pwrmin
	mov	temp, pwrhrL
	rcall	bcdinc
	mov	pwrhrL, temp
	cpi	temp, 0xA0
	brlo	incpwr_end
	
	;100 hours
	clr	pwrhrL
	inc	pwrhrH
	cpi	pwrhrH, 0x0A
	brlo	incpwr_end
	;overflow
	ldi	pwrhrH, 0x19
	ldi	pwrhrL, 0x99
	ldi	pwrsec, 0x59
	ldi	pwrmin, 0x59
incpwr_end:

;---------------------------------------
; return to main program
;---------------------------------------

ovf0_end:
	pop	temp
	out	SREG, temp
	pop	temp2
	pop	temp
	reti
	
;===============================================================================

icp1:
	;Timer 1 Input Capture Event
	;push	temp --> moved to beginning of code segment
	push	temp2
	in	temp, SREG
	push	temp
	
;---------------------------------------
; DCF77 reception
;---------------------------------------
	
	;reset timer
	ldi	temp, 0
	out	TCNT1H, temp
	out	TCNT1L, temp
	
	;get time since last DCF77 signal edge
	in	temp, ICR1L
	in	temp2, ICR1H
	cpi	temp2, 0
	breq	PC+2
	ldi	temp, 0xFF
	in	temp2, TCCR1B
	sbrc	temp2, ICES1
	rjmp	dcf_posedge
	
	;negative edge: one bit received
	ori	temp2, 1<<ICES1 ;next edge will be positive
	out	TCCR1B, temp2
	sbrc	dcfflags, StopDCF ;ignore if reception stopped or disabled
	rjmp	dcf_end
	sbrc	dcfflags, DCFdisabled
	rjmp	dcf_end
	cpi	temp, 9
	brlo	dcf_err_bitlen
	cpi	temp, 17
	brlo	dcf_bit_zero ;75ms <= bit length <= 125 ms  -->  bit = 0
	cpi	temp, 22
	brlo	dcf_err_bitlen
	cpi	temp, 30
	brlo	dcf_bit_one ;175ms <= bit length <= 225 ms  -->  bit = 1
	
dcf_err_bitlen:
	;invalid bit length, stop reception for this telegram
	andi	dcfflags, ~ErrMask
	ori	dcfflags, 1<<ErrBitLength | 1<<StopDCF
	rjmp	dcf_end
	
dcf_bit_zero:
	clc ;clear bit
	rjmp	dcf_bit_save
dcf_bit_one:
	sec ;set bit
dcf_bit_save:
	ror	dcf5
	ror	dcf4
	ror	dcf3
	ror	dcf2
	ror	dcf1
	inc	dcfbits
	cpi	dcfbits, 60 ;stop reception if telegram contains more than 59 bits
	brlo	dcf_end
	andi	dcfflags, ~ErrMask
	ori	dcfflags, 1<<ErrBitCount | 1<<StopDCF
	rjmp	dcf_end

dcf_posedge:
	;positive edge: beginning of a new bit
	andi	temp2, ~(1<<ICES1) ;next edge will be negative
	out	TCCR1B, temp2
	sbrc	dcfflags, DCFdisabled ;ignore if reception disabled
	rjmp	dcf_end
	cpi	temp, 192
	brsh	dcf_complete ;pause >= 1.5s: telegram complete
	
dcf_end:
	pop	temp
	out	SREG, temp
	pop	temp2
	pop	temp
	reti

;---------------------------------------
; DCF77 decoding
;---------------------------------------

dcf_complete:
	;end of telegram / new telegram
	mov	temp, dcfflags
	mov	temp2, dcfbits
	andi	dcfflags, ~(ErrMask | 1<<StopDCF)
	ldi	dcfbits, 0
	sbrc	temp, StopDCF ;don't decode if DCF77 reception was stopped
	rjmp	dcf_end
	cpi	temp2, 59 ;check if correct number of bits were received
	breq	dcf_decode
	ori	dcfflags, 1<<ErrBitCount
	rjmp	dcf_end
	
dcf_decode:
	;check startbit
	sbrs	dcf1, 1
	rjmp	dcf_err_startbit
	
	;check date parity
	ldi	temp2, 0
	mov	temp, dcf3
	andi	temp, 0xFE
	rcall	parity
	mov	temp, dcf4
	rcall	parity
	mov	temp, dcf5
	rcall	parity
	sbrc	temp2, 0
	rjmp	dcf_err_parity
	
	;check hour parity
	ldi	temp2, 0
	mov	temp, dcf2
	andi	temp, 0xFC
	rcall	parity
	sbrc	dcf3, 0
	inc	temp2
	sbrc	temp2, 0
	rjmp	dcf_err_parity
	
	;check minute parity
	ldi	temp2, 0
	mov	temp, dcf1
	andi	temp, 0xFC
	rcall	parity
	mov	temp, dcf2
	andi	temp, 0x03
	rcall	parity
	sbrc	temp2, 0
	rjmp	dcf_err_parity

	;save minutes
	lsr	dcf2
	ror	dcf1
	lsr	dcf1
	mov	minute, dcf1
	
	;save hours
	lsr	dcf2
	mov	hour, dcf2
	
	;save day
	lsl	dcf3 ;align day
	rol	dcf4 ;align month[5] weekday[3]
	rol	dcf5 ;align year
	lsr	dcf3
	lsr	dcf3
	mov	day, dcf3
	
	;save weekday
	mov	wkday, dcf4
	ldi	temp, 0x07
	and	wkday, temp
	
	;save month
	lsr	dcf4
	lsr	dcf4
	lsr	dcf4
	mov	month, dcf4
	
	;save year
	mov	year, dcf5
	
	;set time & date ok, disable DCF77 reception and power off receiver
	clr	second
	ori	dcfflags, 1<<TimeOK | 1<<DCFdisabled | 1<<StopDCF | 1<<DCFTime
	cbi	DCFPORT, DCF_Power
	rjmp	dcf_end
	
dcf_err_parity:
	;set parity error
	ori	dcfflags, 1<<ErrParity
	rjmp	dcf_end

dcf_err_startbit:
	;set start-bit error
	ori	dcfflags, 1<<ErrStartBit
	rjmp	dcf_end

;===============================================================================

ovf1:
	;Timer 1 overflow
	
;===============================================================================

start:
	;set stackpointer
	;ldi	temp, LOW(RAMEND) --> moved to beginning of code segment
	out	SPL, temp
	
	;set ports
	ldi	temp, 0x02
	out	DDRA, temp
	ldi	temp, 0x00
	out	PORTA, temp
	ldi	temp, 0xF0
	out	DDRB, temp
	ldi	temp, 0x0F
	out	PORTB, temp
	ldi	temp, 0x37
	out	DDRD, temp
	ldi	temp, 0x40
	out	PORTD, temp
	
	;disable analog comparator
	ldi	temp, 1<<ACD
	out	ACSR, temp
	
	;init Timers 0 and 1
	ldi	temp, 1<<WGM01 ;Timer 0: CTC mode
	out	TCCR0A, temp
	ldi	temp, 0x05 ;Timer 0: Clk/1024
	out	TCCR0B, temp
	ldi	temp, 31 ;Timer 0: count from 0 to 31 (divide clock by 32)
	out	OCR0A, temp
	ldi	temp, 1<<ICNC1 | 1<<ICES1 | 0x04 ;Timer 1: Clk/256
	out	TCCR1B, temp
	ldi	temp, 1<<ICIE1 | 1<<TOIE1 | 1<<OCIE0A
	out	TIMSK, temp
	
	;init external interrupts
	ldi	temp, 1<<ISC10 ;INT1: trigger on both edges
	out	MCUCR, temp
	ldi	temp, 0x03 ;Pin Change Interrupt for PB0 and PB1
	out	PCMSK, temp
	ldi	temp, 1<<INT1 | 1<<PCIE ;enable INT1 and Pin Change Interrupt
	out	GIMSK, temp
	
	;init registers
	clr	second
	clr	minute
	clr	hour
	ldi	temp, 0x01
	mov	day, temp
	mov	wkday, temp
	mov	month, temp
	mov	year, temp
	clr	dcf5
	ldi	dcfflags, 1<<StopDCF
	ldi	pwrsec, INVALID
	ldi	keys, 0x00
	clr	oldkeys
	
	;display splash screen
	rcall	lcd_on
	ldi	ZL, LOW(strStart*2)
	ldi	ZH, HIGH(strStart*2)
	rcall	flash2lcd
	ldi	count, 100
	rcall	longwait
	
	;enable interrupts, switch on DCF77 receiver
	sei
	sbi	DCFPORT, DCF_Power
	
;-------------------------------------------------------------------------------

	;display DCF77 reception information until correct time received
wait_dcf:
	;check if user wants to set time manually (both keys, long press)
	cpi	bothkeys, 3
	brne	PC+2
	rcall	settime
	
	;check if TCCR1B has changed (ICES1 toggles on each DCF77 signal edge)
	in	temp2, TCCR1B
	cp	temp2, temp
	breq	wait_dcf
	
	;update DCF77 information on the LCD
	sbrc	dcfflags, TimeOK
	rjmp	main
	ldi	ZL, LOW(strDCFInfo*2) ;static text elements
	ldi	ZH, HIGH(strDCFInfo*2)
	rcall	flash2lcd
	ldi	temp, 0x80 ;bit count
	rcall	lcdcmd
	mov	temp, dcfbits
	rcall	bin2lcd
	ldi	temp, 0x8E ;last bit value
	rcall	lcdcmd
	ldi	temp, '0'
	sbrc	dcf5, 7
	ldi	temp, '1'
	rcall	lcdchr
	ldi	temp, 0xC4 ;last error
	rcall	lcdcmd
	ldi	temp, '-'
	sbrc	dcfflags, ErrBitCount
	ldi	temp, 'C'
	sbrc	dcfflags, ErrBitLength
	ldi	temp, 'L'
	sbrc	dcfflags, ErrParity
	ldi	temp, 'P'
	sbrc	dcfflags, ErrStartBit
	ldi	temp, 'S'
	rcall	lcdchr
	ldi	ZL, LOW(strStop*2)
	ldi	ZH, HIGH(strStop*2)
	sbrs	dcfflags, StopDCF
	adiw	ZH:ZL, (strSpace-strStop)*2
	ldi	temp, 0xC6
	rcall	str2lcd
	ldi	temp, 0xCE ;DCF77 input state
	rcall	lcdcmd
	ldi	temp, '_'
	in	temp2, TCCR1B
	sbrs	temp2, ICES1
	ldi	temp, '-'
	rcall	lcdchr
	in	temp, TCCR1B
	rjmp	wait_dcf
	
;-------------------------------------------------------------------------------

main:
	rcall	lcdclear
	ldi	lcdtime, LCD_TIMEOUT
	
	
	
mainloop:
	;go to sleep
	in	temp, MCUCR
	ori	temp, 1<<SE
	andi	temp, ~(1<<SM) ;select Idle Mode
	out	MCUCR, temp
	sleep
	andi	temp, ~(1<<SE)
	out	MCUCR, temp



main_update:
	;check keys
	sbic	LCDCtrl, LCD_Power
	rjmp	checkkeys_individual
	
	;LCD is off: only enable LCD on keypress, do nothing else
	ldi	dispaddr, INVALID
	;check if user wants to set time manually (both keys, long press)
	cpi	bothkeys, 3
	brlo	PC+2
	rcall	settime
	;check if any keys were pressed
	mov	temp, keys
	andi	temp, Keypresses
	breq	checkkeys_end ;no keys pressed
	andi	keys, ~(Keypresses) ;clear all keypresses
	ldi	lcdtime, LCD_TIMEOUT
	rcall	lcd_on
	rjmp	checkkeys_end
	
checkkeys_individual:
	;LCD is on: check keys individually
	sbrc	keys, kBothPress
	rjmp	key_both
	sbrc	keys, kUpPress
	rjmp	key_next
	sbrc	keys, kDownPress
	rjmp	key_prev
checkkeys_end:
	
	
	
	;update display (if enabled and no power outage information is shown)
	sbis	LCDCtrl, LCD_Power
	rjmp	mainloop ;don't display anything if LCD is switched off
	;check if LCD timeout reached
	cpi	lcdtime, 0
	brne	lcdcheck_end
	rcall	lcd_off
	rjmp	mainloop
lcdcheck_end:

	;check if time/date should be displayed
	cpi	dispaddr, INVALID
	brne	mainloop
	
	;display time, date and DCF77 information
	rcall	time2lcd
	ldi	ZL, LOW(strDCF*2)
	ldi	ZH, HIGH(strDCF*2)
	sbrs	dcfflags, DCFTime
	adiw	ZH:ZL, (strSpace-StrDCF)*2
	ldi	temp, 0x8D
	rcall	str2lcd
	ldi	ZL, LOW(strRx*2)
	ldi	ZH, HIGH(strRx*2)
	sbrc	dcfflags, DCFdisabled
	adiw	ZH:ZL, (strSpace-StrRx)*2
	ldi	temp, 0xCE
	rcall	str2lcd
	
	rjmp	mainloop
	
;-------------------------------------------------------------------------------

settime:
	;set clock time and date
	sbrc	dcfflags, TimeOK ;clear time/date if still unset
	rjmp	settime_clearend
	clr	year
	ldi	temp, 1
	mov	month, temp
	mov	day, temp
	mov	wkday, temp
	clr	hour
	clr	minute
	clr	second
settime_clearend:
	andi	keys, ~(Keypresses) ;clear all keypresses
	
	;display old time/date settings
	andi	dcfflags, ~(1<<TimeOK) ;don't increment time while setting
	rcall	lcd_on
	rcall	time2lcd
	ldi	temp, 0x0E ;cursor on
	rcall	lcdcmd
	ldi	ZL, LOW(strSet*2)
	ldi	ZH, HIGH(strSet*2)
	ldi	temp, 0x8D
	rcall	str2lcd
	
	;Set Date
	ldi	maxval, 0x99	;year
	clr	minval ;minval = 0x00
	mov	count, year
	ldi	dispaddr, 0xC2
	rcall	setnum
	mov	year, count
	ldi	maxval, 0x12	;month
	inc	minval ;minval = 0x01
	mov	count, month
	ldi	dispaddr, 0xC5
	rcall	setnum
	mov	month, count
	rcall	monthlength	;day
	mov	maxval, temp
	cp	temp, day
	brsh	PC+2
	mov	day, temp
	mov	count, day
	ldi	dispaddr, 0xC8
	rcall	setnum
	mov	day, count
	
	;Set Weekday
	mov	count, wkday
settime_weekday:
	;update LCD and move cursor to weekday
	rcall	time2lcd
	ldi	temp, 0x80
	rcall	lcdcmd
settime_weekday_wait:
	;wait for keypress
	sbrc	keys, kDownPress
	rjmp	settime_weekday_end
	sbrs	keys, kUpPress
	rjmp	settime_weekday_wait
	;increment value
	andi	keys, ~(Keypresses) ;clear all keypresses
	inc	count
	sbrc	count, 3 ;count == 8?
	ldi	count, 1
	mov	wkday, count
	;display new weekday
	rcall	weekday2lcd
	rjmp	settime_weekday
settime_weekday_end:
	andi	keys, ~(Keypresses) ;clear all keypresses
	
	;Set Time
	ldi	maxval, 0x23	;hour
	clr	minval ;minval = 0x00
	mov	count, hour
	ldi	dispaddr, 0x83
	rcall	setnum
	mov	hour, count
	ldi	maxval, 0x59	;minute
	mov	count, minute
	ldi	dispaddr, 0x86
	rcall	setnum
	mov	minute, count
	mov	count, second	;second
	ldi	dispaddr, 0x89
	rcall	setnum
	mov	second, count
	
	;return
	andi	keys, ~(Keypresses) ;clear all keypresses
	ldi	temp, 0xC0 ;cursor off
	rcall	lcdcmd
	rcall	lcdclear
	andi	dcfflags, ~(1<<DCFTime)
	ori	dcfflags, 1<<TimeOK | 1<<DCFdisabled
	cbi	DCFPORT, DCF_Power
	ldi	dispaddr, INVALID
	ret
	
;--------------------
	
setnum:
	;select packed BCD number 'count' from interval [minval, maxval]
	;update LCD and move cursor to selected value
	mov	temp, dispaddr
	rcall	lcdcmd
	mov	temp, count
	rcall	bcd2lcd
	mov	temp, dispaddr
	rcall	lcdcmd
	
setnum_keywait:
	;wait for keypress
	sbrc	keys, kDownPress
	rjmp	setnum_done
	sbrs	keys, kUpPress
	rjmp	setnum_keywait
	
	;increment value
	andi	keys, ~(Keypresses) ;clear all keypresses
	mov	temp, count
	rcall	bcdinc
	cp	maxval, temp ;wrap around
	brsh	PC+2
	mov	temp, minval
	mov	count, temp
	rjmp	setnum
	
setnum_done:
	andi	keys, ~(Keypresses) ;clear all keypresses
	ret
	
;-------------------

key_both:
	;both keys pressed at the same time: switch off LCD immediately
	andi	keys, ~(Keypresses) ;clear all keypresses
	rcall	lcd_off ;LCD off
	rjmp	main_update

;--------------------
	
key_prev:
	;"down" key pressed: display previous power outage information
	andi	keys, ~(Keypresses) ;clear all keypresses
	cpi	dispaddr, INVALID
	breq	findfirst
	subi	dispaddr, 10
	brcc	PC+2
	ldi	dispaddr, 110
	mov	ZL, dispaddr
	rcall	eeprom_read
	cpi	temp, EMPTY
	brne	outage_disp
	;end of list reached, undo changes to 'dispaddr'
	subi	dispaddr, -10
	cpi	dispaddr, 120
	brlo	PC+2
	ldi	dispaddr, 0
	rjmp	main_update
	
;--------------------

key_next:
	;"up" key pressed: display next power outage information
	andi	keys, ~(Keypresses) ;clear all keypresses
	cpi	dispaddr, INVALID
	breq	findfirst
	subi	dispaddr, -10
	cpi	dispaddr, 120
	brlo	PC+2
	ldi	dispaddr, 0
	mov	ZL, dispaddr
	rcall	eeprom_read
	cpi	temp, EMPTY
	brne	outage_disp
	;end of list reached, undo changes to 'dispaddr'
	subi	dispaddr, 10
	brcc	PC+2
	ldi	dispaddr, 110
	rjmp	main_update

;--------------------

findfirst:
	;find most recent power outage information in EEPROM
	rcall	eeprom_findfree
	subi	ZL, 10
	brcc	PC+2
	ldi	ZL, 110
	rcall	eeprom_read
	cpi	temp, EMPTY
	breq	findfirst_empty ;EEPROM is empty
	mov	dispaddr, ZL
	rjmp	outage_disp

findfirst_empty:
	;display "empty" message
	ldi	ZL, LOW(strEmpty*2)
	ldi	ZH, HIGH(strEmpty*2)
	rcall	flash2lcd
	ldi	count, MESSAGE_TIME
	rcall	longwait
	rcall	lcdclear
	ldi	dispaddr, INVALID
	rjmp	main_update

;--------------------

outage_disp:
	;display selected power outage information
	rcall	lcdclear
	mov	ZL, dispaddr
	ldi	temp, 0xC0 ;year
	rcall	lcdcmd
	ldi	temp, '2'
	rcall	lcdchr
	ldi	temp, '0'
	rcall	lcdchr
	rcall	eeprom2lcd
	ldi	temp, 0x80 ;month/day
	rcall	lcdcmd
	rcall	eeprom2lcd
	ldi	temp, '/'
	rcall	lcdchr
	rcall	eeprom2lcd
	ldi	temp, ' '
	rcall	lcdchr
	rcall	eeprom2lcd ;time
	ldi	temp, ':'
	rcall	lcdchr
	rcall	eeprom2lcd
	ldi	temp, ':'
	rcall	lcdchr
	rcall	eeprom2lcd
	ldi	temp, 0xC5 ;duration
	rcall	lcdcmd
	ldi	temp, 'd'
	rcall	lcdchr
	ldi	temp, '='
	rcall	lcdchr
	rcall	eeprom_read
	inc	ZL
	andi	temp, 0x0F
	ori	temp, 0x30
	rcall	lcdchr
	rcall	eeprom2lcd
	ldi	temp, ':'
	rcall	lcdchr
	rcall	eeprom2lcd
	ldi	temp, ':'
	rcall	lcdchr
	rcall	eeprom2lcd
	rjmp	main_update
	
;===============================================================================

time2lcd:
	;display time and date on the LCD

	;Time
	ldi	temp, 0x83
	rcall	lcdcmd
	mov	temp, hour
	rcall	bcd2lcd
	ldi	temp, ':'
	rcall	lcdchr
	mov	temp, minute
	rcall	bcd2lcd
	ldi	temp, ':'
	rcall	lcdchr
	mov	temp, second
	rcall	bcd2lcd
	mov	temp, second
	
	;Date
	ldi	temp, 0xC0
	rcall	lcdcmd
	ldi	temp, '2'
	rcall	lcdchr
	ldi	temp, '0'
	rcall	lcdchr
	mov	temp, year
	rcall	bcd2lcd
	ldi	temp, '-'
	rcall	lcdchr
	mov	temp, month
	rcall	bcd2lcd
	ldi	temp, '-'
	rcall	lcdchr
	mov	temp, day
	rcall	bcd2lcd

weekday2lcd:
	;Weekday
	ldi	temp, 0x80
	rcall	lcdcmd
	ldi	ZL, LOW(weekdays*2-2)
	ldi	ZH, HIGH(weekdays*2-2)
	mov	temp, wkday
	lsl	temp
	add	ZL, temp
	ldi	temp, 0
	adc	ZH, temp
	lpm	temp, Z+
	rcall	lcdchr
	lpm	temp, Z+
	rcall	lcdchr
	ret

;--------------------

lcd_on:
	;enable LCD
	sbi	LCDCtrl, LCD_Power
	ldi	temp, 255
	rcall	wait
	rcall	lcdinit
	ret
	
;--------------------

lcd_off:
	;disable LCD
	cbi	LCDCtrl, LCD_RS
	cbi	LCDCtrl, LCD_E
	ldi	temp, 0x0F
	out	LCDData, temp
	cbi	LCDCtrl, LCD_Power
	ret

;--------------------

flash2lcd: ;copy 32 bytes from Flash (ZH:ZL) to LCD (2 lines, 16 chars each)
	ldi	temp, 0x80
	rcall	lcdcmd
	ldi	count, 16
flash2lcd_loop1:
	lpm	temp, Z+
	rcall	lcdchr
	dec	count
	brne	flash2lcd_loop1
	ldi	temp, 0xC0
	rcall	lcdcmd
	ldi	count, 16
flash2lcd_loop2:
	lpm	temp, Z+
	rcall	lcdchr
	dec	count
	brne	flash2lcd_loop2
	ret

;--------------------

eeprom2lcd:
	;display packed BCD value from EEPROM and increment EEPROM address
	rcall	eeprom_read
	inc	ZL
	rjmp	bcd2lcd

;--------------------

str2lcd:
	;copy string at ZH:ZL to LCD, stop at char 0x00
	rcall	lcdcmd
str2lcd_loop:
	lpm	temp, Z+
	cpi	temp, 0
	breq	str2lcd_end
	rcall	lcdchr
	rjmp	str2lcd_loop
str2lcd_end:
	ret
	
;--------------------

bin2lcd:
	;display binary value
	;convert binary to packed BCD
	ldi	temp2, 0
bin2lcd_loop:
	subi	temp2, -0x10
	subi	temp, 10
	brcc	bin2lcd_loop
	subi	temp2, 0x10
	subi	temp, -10
	add	temp, temp2
	;(fall through to bcd2lcd)
	
;--------------------


bcd2lcd:
	;display packed BCD value
	push	temp
	swap	temp
	andi	temp, 0x0F
	ori	temp, 0x30
	rcall	lcdchr
	pop	temp
	andi	temp, 0x0F
	ori	temp, 0x30
	rcall	lcdchr
	ret
	
;--------------------

lcdinit: ;initialize LCD (4 bit mode)
	ldi	temp, 150
	rcall	wait
	ldi	temp, 0x30
	rcall	lcdcmd8bit
	ldi	temp, 41
	rcall	wait
	ldi	temp, 0x30
	rcall	lcdcmd8bit
	ldi	temp, 0x30
	rcall	lcdcmd8bit
	ldi	temp, 0x20
	rcall	lcdcmd8bit
	ldi	temp, 0x2A
	rcall	lcdcmd
	ldi	temp, 0x0C
	rcall	lcdcmd
	ldi	temp, 0x06
	rcall	lcdcmd
	ldi	temp, 0x02
	rcall	lcdcmd
	ldi	temp, 15
	rcall	wait
	;user defined characters
	ldi	temp, 0x40
	rcall	lcdcmd
	ldi	ZL, LOW(charLemon*2)
	ldi	ZH, HIGH(charLemon*2)
	ldi	count, 64
lcdinit_userchars:
	lpm	temp, Z+
	rcall	lcdchr
	dec	count
	brne	lcdinit_userchars
	;(fall through to lcdclear)

;--------------------

lcdclear: ;clear LCD
	ldi	temp, 0x01
	rcall	lcdcmd
	ldi	temp, 15
	rcall	wait
	ret

;--------------------

lcdcmd: ;send a command to the LCD
	cbi	LCDCtrl, LCD_RS
	mov	temp2, temp
	ori	temp2, 0x0F
	out	LCDData, temp2
	nop
	sbi	LCDCtrl, LCD_E
	nop
	cbi	LCDCtrl, LCD_E
	swap	temp
	ori	temp, 0x0F
	out	LCDData, temp
	nop
	sbi	LCDCtrl, LCD_E
	nop
	cbi	LCDCtrl, LCD_E
	ret

;--------------------

lcdchr: ;send a character to the LCD
	sbi	LCDCtrl, LCD_RS
	mov	temp2, temp
	ori	temp2, 0x0F
	out	LCDData, temp2
	nop
	sbi	LCDCtrl, LCD_E
	nop
	cbi	LCDCtrl, LCD_E
	swap	temp
	ori	temp, 0x0F
	out	LCDData, temp
	nop
	sbi	LCDCtrl, LCD_E
	nop
	cbi	LCDCtrl, LCD_E
	ret

;--------------------

lcdcmd8bit: ;send a command to the LCD in 8-Bit-Mode (only high nibble is sent)
	cbi	LCDCtrl, LCD_RS
	ori	temp, 0x0F
	out	LCDData, temp
	nop
	sbi	LCDCtrl, LCD_E
	nop
	cbi	LCDCtrl, LCD_E
	ret
	
;--------------------

longwait:
	ldi	temp, 0
	rcall	wait
	dec	count
	brne	longwait
	ret

;--------------------

wait: ;wait for about temp * 150us (100us needed, add 50us for 3V operation)
	rjmp	PC+1
	dec	temp
	brne	wait
	ret
	
;===============================================================================

;---------------------------------------
; EEPROM functions
;---------------------------------------

eeprom_write:
	;write a byte to EEPROM and increment address
	sbic	EECR, EEPE
	rjmp	eeprom_write
	out	EEAR, ZL
	out	EEDR, temp
	sbi	EECR, EEMPE
	sbi	EECR, EEPE
	inc	ZL
	andi	ZL, 0x7F
	ret

;--------------------

eeprom_read:
	;read a byte from EEPROM
	sbic	EECR, EEPE
	rjmp	eeprom_read
	out	EEAR, ZL
	sbi	EECR, EERE
	in	temp, EEDR
	ret

;--------------------

eeprom_findfree:
	;find first free EEPROM slot
	ldi	ZL, 0
eeprom_findfree_loop:
	rcall	eeprom_read
	cpi	temp, EMPTY
	breq	eeprom_findfree_end
	subi	ZL, -10
	cpi	ZL, 120
	brlo	eeprom_findfree_loop
	;no free slot found: something went wrong, clear first slot
	ldi	temp, EMPTY
	ldi	ZL, 0
	rcall	eeprom_write
	ldi	ZL, 0
eeprom_findfree_end:
	ret

;---------------------------------------
; DCF77/Clock helper functions
;---------------------------------------

parity:
	;calculate parity (add to temp2)
	sbrc	temp, 0
	inc	temp2
	lsr	temp
	brne	parity
	ret
	
;--------------------

bcdinc:
	;increment packed bcd number
	subi	temp, -7 ;increment and add 6
	brhc	PC+2 ;half-carry is inverted when using "subi rh, -k" instead of add!
	subi	temp, 6
	ret
	
;--------------------

monthlength:
	;calculate length of month in days
	ldi	temp, 0x30
	mov	temp2, month
	cpi	temp2, 0x08
	brlo	PC+2
	inc	temp2
	sbrc	temp2, 0
	ldi	temp, 0x31
	cpi	temp2, 0x02
	brne	monthlength_end
	ldi	temp, 0x28
	mov	temp2, year
	andi	temp2, 0x03
	brne	monthlength_end
	ldi	temp, 0x29
	rjmp	monthlength_end
monthlength_end:
	ret

;--------------------

bcd2bin:
	;convert packed BCD to binary
	mov	temp2, temp
	andi	temp, 0x0F
	swap	temp2
	andi	temp2, 0x0F
	lsl	temp2
	add	temp, temp2 ;temp += temp2*2
	lsl	temp2
	lsl	temp2
	add	temp, temp2 ;temp += temp2*8
	ret
